from typing import Literal
from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.grants.volumegrant import VolumeGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Volume(BaseModel, PulumiResource, TerraformResource):
    """
    Volumes are Unity Catalog objects representing a logical volume of storage
    in a cloud object storage location. Volumes provide capabilities for
    accessing, storing, governing, and organizing files. While tables provide
    governance over tabular datasets, volumes add governance over non-tabular
    datasets. You can use volumes to store and access files in any format,
    including structured, semi-structured, and unstructured data.

    Examples
    --------
    ```py
    from laktory import models

    volume = models.resources.databricks.Volume(
        name="landing",
        catalog_name="dev",
        comment="Landing zone for raw data",
        schema_name="sources",
        volume_type="EXTERNAL",
        storage_location="abfss://landing@lakehouse-storage.dfs.core.windows.net/",
        grants=[
            {"principal": "account users", "privileges": ["READ_VOLUME"]},
            {"principal": "role-metastore-admins", "privileges": ["WRITE_VOLUME"]},
        ],
    )
    print(volume.full_name)
    # > dev.sources.landing
    print(volume.parent_full_name)
    # > dev.sources
    ```

    References
    ----------

    * [Databricks Volume](https://docs.databricks.com/en/sql/language-manual/sql-ref-volumes.html)
    * [Pulumi Databricks Volume](https://www.pulumi.com/registry/packages/databricks/api-docs/volume/)
    """

    name: str = Field(..., description="Name of the volume")
    catalog_name: str = Field(
        None, description="Name of the catalog storing the volume"
    )
    comment: str = Field(None, description="Text description of the volume")
    grant: Union[VolumeGrant, list[VolumeGrant]] = Field(
        None,
        description="""
    Grant(s) operating on the Volume and authoritative for a specific principal. Other principals within the grants are 
    preserved. Mutually exclusive with `grants`. 
    """,
    )
    grants: list[VolumeGrant] = Field(
        None,
        description="""
    Grants operating on the Volume and authoritative for all principals. Replaces any existing grants defined inside or
    outside of Laktory. Mutually exclusive with `grant`.
    """,
    )
    schema_name: str = Field(None, description="Name of the schema storing the volume")
    storage_location: str = Field(
        None,
        description="Path inside an External Location. Only used for EXTERNAL Volumes.",
    )
    volume_type: Literal["MANAGED", "EXTERNAL"] = Field(
        "MANAGED",
        description="""
    Type of volume. A managed volume is a Unity Catalog-governed storage volume created within the default storage 
    location of the containing schema. An external volume is a Unity Catalog-governed storage volume registered against
    a directory within an external location.
    """,
    )

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Schema full name `{catalog_name}.{schema_name}`"""
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        """Volume full name `{catalog_name}.{schema_name}.{volume_name}`"""
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Table full name (catalog.schema.volume)"""
        return self.full_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - volume grants
        """
        resources = []

        # Volume grants
        resources += self.get_grants_additional_resources(
            object={"volume": f"${{resources.{self.resource_name}.id}}"}
        )
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Volume"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["grant", "grants"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_volume"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Any
from typing import Literal

import narwhals as nw
from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.datasources.basedatasource import BaseDataSource

logger = get_logger(__name__)

AnyFrame = nw.DataFrame | nw.LazyFrame


class DataFrameDataSource(BaseDataSource):
    """
    Data source using in-memory DataFrame.

    Examples
    ---------
    From data with PySpark backend.
    ```python
    import laktory as lk

    data = {
        "x": [0, 1],
        "y": ["a", "b"],
    }

    # From data using PySpark
    source = lk.models.DataFrameDataSource(
        data=data,
        dataframe_backend="PYSPARK",
    )
    df = source.read()
    print(df.collect(backend="pandas"))
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    |        x  y      |
    |     0  0  a      |
    |     1  1  b      |
    └──────────────────┘
    '''
    ```

    From Polars DataFrame
    ```python
    import polars as pl

    import laktory as lk

    data = {
        "x": [0, 1],
        "y": ["a", "b"],
    }

    source = lk.models.DataFrameDataSource(
        df=pl.DataFrame(data),
    )
    df = source.read()
    print(df)
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    |    | x | y |     |
    |    |---|---|     |
    |    | 0 | a |     |
    |    | 1 | b |     |
    └──────────────────┘
    '''
    ```
    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/sourcessinks/)
    """

    data: dict[str, list[Any]] | list[dict[str, Any]] = Field(
        None, description="Serialized data used to build source"
    )
    df: Any = Field(None, description="DataFrame object acting as source")
    type: Literal["DATAFRAME"] = Field(
        "DATAFRAME",
        frozen=True,
        description="Source Type",
    )

    @model_validator(mode="after")
    def validate_input(self) -> Any:
        if self.df is None and self.data is None:
            raise ValueError("Either `data` or `df` must be provided.")

        if self.df is not None and self.data is not None:
            raise ValueError("Only `data` or `df` can be provided.")

        if self.df is not None:
            if not isinstance(self.df, (nw.DataFrame, nw.LazyFrame)):
                self.df = nw.from_native(self.df)

            with self.validate_assignment_disabled():
                self.dataframe_backend_ = DataFrameBackends.from_nw_implementation(
                    self.df.implementation
                )

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        n = 4

        # Build Schema-like
        if self.df is not None:
            d = self.df.schema
        elif isinstance(self.data, dict):
            d = {k: str(type(v)) for k, v in self.data.items()}
        elif isinstance(self.data, list):
            d = {k: str(type(v)) for k, v in self.data[0].items()}
        else:
            d = {}

        # Build id
        _id = "DataFrame[" + ", ".join([f"{k}: {v}" for k, v in d.items()][:n]) + "]"
        if len(d) >= n:
            _id = _id.replace("]", "...]")

        if isinstance(self, nw.LazyFrame):
            _id = _id.replace("DataFrame", "LazyFrame")

        return _id

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self) -> nw.LazyFrame:
        from laktory import get_spark_session

        spark = get_spark_session()

        if self.df is not None:
            return nw.from_native(self.df)

        data = self.data
        if isinstance(data, dict):
            data = [dict(zip(data.keys(), values)) for values in zip(*data.values())]

        df = spark.createDataFrame(data)

        return nw.from_native(df)

    def _read_polars(self) -> AnyFrame:
        if self.df is not None:
            return nw.from_native(self.df)

        import polars as pl

        df = pl.LazyFrame(self.data)

        return nw.from_native(df)



from typing import Literal

from pydantic import Field

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel

logger = get_logger(__name__)


class DataQualityCheck(BaseModel):
    """
    A Data Quality Check is the result of an expectation compared to a dataset.

    Examples
    --------
    ```py
    from laktory import models

    check = models.DataQualityCheck(
        rows_count=10,
        fails_count=2,
        status="FAIL",
    )
    print(check)
    # > variables={} fails_count=2 rows_count=10 status='FAIL'

    check = models.DataQualityCheck(
        rows_count=10,
        fails_count=2,
        status="PASS",
    )
    print(check)
    # > variables={} fails_count=2 rows_count=10 status='PASS'
    ```
    References
    ----------
    * [Data Quality](https://www.laktory.ai/concepts/dataquality/)
    """

    fails_count: int = Field(
        None, description="Number of rows not meeting the expectation."
    )
    rows_count: int = Field(None, description="Total number of rows in dataset.")
    status: Literal["PASS", "FAIL"] = Field(
        ...,
        description="Result of comparison, considering the expectation criteria and tolerances.",
    )

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def failure_rate(self):
        """
        Ratio of the number of rows not meeting the expectation to the total number of rows in the dataframe. Returns
        `None` when the type is `"AGGREGATE"`.
        """
        # Empty DataFrame
        if self.rows_count == 0:
            return 0

        # Aggregate Expectation
        if self.fails_count is None:
            return None

        return self.fails_count / self.rows_count



from typing import Any
from typing import Literal

from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.basemodel import BaseModel
from laktory.typing import AnyFrame

logger = get_logger(__name__)

SUPPORTED_BACKENDS = [DataFrameBackends.PYSPARK]


class DataSinkMergeCDCOptions(BaseModel):
    """
    Options for merging a change data capture (CDC).

    They are also used to build the target using `apply_changes` method when
    using Databricks DLT.

    Examples
    --------
    ```py
    from laktory import models

    df = spark.createDataFrame(
        [
            {"id": 1, "value": 3.0},
            {"id": 2, "value": 2.3},
            {"id": 3, "value": 7.7},
        ]
    )

    sink = models.FileDataSink(
        path="./my_table/",
        format="DELTA",
        mode="MERGE",
        merge_cdc_options={
            "scd_type": 1,
            "primary_keys": ["id"],
        },
    )
    # sink.write(df)
    ```

    References
    ----------
    * [Change Data Capture](https://www.laktory.ai/concepts/cdc/)
    * [How to Implement SCD 2 using Delta Table](https://iterationinsights.com/article/how-to-implement-slowly-changing-dimensions-scd-type-2-using-delta-table/)
    * [Change Data Capture with Databricks DLT](https://docs.databricks.com/en/delta-live-tables/python-ref.html#change-data-capture-with-python-in-delta-live-tables)
    """

    # apply_as_truncates: Union[str, None] = None
    delete_where: str = Field(
        None,
        description="Specifies when a CDC event should be treated as a DELETE rather than an upsert.",
    )
    end_at_column_name: str = Field(
        "__end_at",
        description="""
        When using SCD type 2, name of the column storing the end time (or sequencing index) during which a row is 
        active. This attribute is not used when using Databricks DLT which does not allow column rename.
        """,
    )
    exclude_columns: list[str] = Field(
        None, description="A subset of columns to exclude in the target table."
    )
    ignore_null_updates: bool = Field(
        False,
        description="""
        Allow ingesting updates containing a subset of the target columns. When a CDC event matches an existing row and
        ignore_null_updates is `True`, columns with a null will retain their existing values in the target. This also 
        applies to nested columns with a value of null. When ignore_null_updates is `False`, existing values will be 
        overwritten with null values.
        """,
    )
    include_columns: list[str] = Field(
        None,
        description="""
        A subset of columns to include in the target table. Use `include_columns` to specify the complete list of 
        columns to include.
        """,
    )
    order_by: str = Field(
        None,
        description="""
        The column name specifying the logical order of CDC events in the source data. Used to handle
        change events that arrive out of order.
        """,
    )
    primary_keys: list[str] = Field(
        None,
        description="""
        The column or combination of columns that uniquely identify a row in the source data. This is used to 
        identify which CDC events apply to specific records in the target table.
        """,
    )
    scd_type: Literal[1, 2] = Field(
        1, description="Whether to store records as SCD type 1 or SCD type 2."
    )
    start_at_column_name: str = Field(
        "__start_at",
        description="""
        When using SCD type 2, name of the column storing the start time (or sequencing index) during which 
        a row is active. This attribute is not used when using Databricks DLT which does not allow column rename.
        """,
    )
    # track_history_columns: Union[list[str], None] = None
    # track_history_except_columns: Union[list[str], None] = None
    _parent: Any = None
    _source_schema: Any = None
    _source_columns: list[str] = None

    @model_validator(mode="after")
    def validate_scd_type(self) -> Any:
        if self.scd_type == 2:
            if self.order_by is None:
                raise ValueError(
                    "SCD Type 2 merge requires specific of `order_by` attribute."
                )

        return self

    @model_validator(mode="after")
    def selected_columns(self) -> Any:
        if self.include_columns and self.exclude_columns:
            raise ValueError(
                "`include_columns` and `exclude_columns` attributes are mutually exclusive."
            )

        if self.exclude_columns and self.order_by:
            if self.order_by in self.exclude_columns:
                raise ValueError(
                    f"`order_by` '{self.order_by}' can't be excluded as it will prevent proper merge of out-of-sequence records."
                )

        return self

    # ----------------------------------------------------------------------- #
    # Sink                                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def sink(self):
        return self._parent

    @property
    def target_name(self):
        from laktory.models.datasinks.tabledatasink import TableDataSink

        if self._parent and isinstance(self._parent, TableDataSink):
            return self._parent.full_name
        return None

    @property
    def target_path(self):
        from laktory.models.datasinks.filedatasink import FileDataSink

        if self._parent and isinstance(self._parent, FileDataSink):
            return self._parent.path
        return None

    @property
    def target_id(self):
        return self.target_name or self.target_path

    # ----------------------------------------------------------------------- #
    # CDC Columns                                                             #
    # ----------------------------------------------------------------------- #

    @property
    def index(self):
        return self.order_by

    @property
    def start_at(self):
        return self.start_at_column_name

    @property
    def end_at(self):
        return self.end_at_column_name

    @property
    def index_fist(self):
        return self.index + "_first"

    @property
    def hash_cols(self):
        return "__hash_cols"

    @property
    def source_columns(self):
        return [f.name for f in self._source_schema]

    @property
    def update_columns(self):
        if self.include_columns:
            columns = [c for c in self.include_columns]
        else:
            columns = [c for c in self.source_columns if c not in self.primary_keys]
            if self.exclude_columns:
                columns = [c for c in columns if c not in self.exclude_columns]
        return columns

    @property
    def extra_columns(self):
        cols = []
        if self.scd_type == 2:
            cols += [self.start_at, self.end_at, self.hash_cols]
        return cols

    @property
    def write_columns(self):
        return self.primary_keys + self.update_columns + self.extra_columns

    @property
    def index_type(self):
        return [
            field.dataType
            for field in self._source_schema.fields
            if field.name == self.index
        ][0]

    @property
    def source_delete_where(self):
        return self._add_alias(self.delete_where)

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    @staticmethod
    def _add_alias(expr, prefix="source"):
        operators = ["=", ">", "<", "!", "*", "+", "-", "/", ","]

        new_expr = expr
        for o in operators:
            new_expr = new_expr.replace(o, " ")

        candidates = [c for c in new_expr.split(" ") if c != ""]

        # Define other exclusions for SQL keywords or constants
        exclusions = [
            "TRUE",
            "FALSE",
            "CASE",
            "WHEN",
            "THEN",
            "SELECT",
            "FROM",
            "WHERE",
            "AND",
            "OR",
            "AS",
            "END",
            "ELSE",
            "NOT",
            "NULL",
        ]

        # Filter out identifiers that are constants or keywords
        column_names = []
        for c in candidates:
            if c.upper() in exclusions:
                continue
            if c.startswith("'"):
                continue
            try:
                float(c)
                continue
            except ValueError:
                pass

            column_names += [c]

        new_expr = expr
        for c in column_names:
            if f"{prefix}." not in c:
                new_expr = new_expr.replace(c, f"{prefix}.{c}")

        return new_expr

    def _init_target(self, source):
        import pyspark.sql.types as T

        spark = source.sparkSession
        logger.info(f"Merge target not found. Creating empty table at {self.target_id}")
        schema = source.select(self.primary_keys + self.update_columns).schema
        if self.scd_type == 2:
            schema.add(T.StructField(self.hash_cols, T.StringType(), True))
            schema.add(T.StructField(self.start_at, self.index_type, True))
            schema.add(T.StructField(self.end_at, self.index_type, True))

        df = spark.createDataFrame(data=[], schema=schema)

        writer = df.write.format("delta").mode("OVERWRITE")
        # if self.sink.cluster_by:
        #     writer = writer.clusterBy(*self.sink.cluster_by)
        if self.target_path:
            writer.save(self.target_path)
        else:
            writer.saveAsTable(self.target_name)

    def _execute(self, source: AnyFrame):
        import pyspark.sql.functions as F
        from delta.tables import DeltaTable
        from pyspark.sql import Window

        spark = source.sparkSession

        logger.info(
            f"Executing merge on {self.target_id} with primary keys {self.primary_keys} and scd type {self.scd_type}"
        )

        if self.delete_where:
            logger.info(f"with delete on {self.delete_where}")

        # Add internal columns
        if self.scd_type == 2:
            source = (
                source.withColumn(self.start_at, F.col(self.index))
                .withColumn(self.end_at, F.lit(None).cast(self.index_type))
                .withColumn(
                    self.hash_cols,
                    F.lit(F.sha2(F.concat_ws("~", *self.update_columns), 256)),
                )
            )

        # Process History
        if self.index:
            w = Window.partitionBy(*self.primary_keys).orderBy(F.desc(self.index))
            source = source.withColumn("_row_number", F.row_number().over(w))
            if self.scd_type == 1:
                # Drop Duplicates
                logger.info(
                    f"Dropping duplicates using {self.primary_keys} and '{self.order_by}' as sequencing index"
                )
                source = source.filter(F.col("_row_number") == 1)
            elif self.scd_type == 2:
                # Assign previous index to ends_at
                w2 = Window.partitionBy(*self.primary_keys)
                source = source.withColumn(self.end_at, F.lag(self.index, 1).over(w))
                source = source.withColumn(self.index_fist, F.min(self.index).over(w2))
            source = source.drop("_row_number")
        else:
            logger.info(f"Dropping duplicates using {self.primary_keys}")
            source = source.drop_duplicates(subset=self.primary_keys)

        # Read target
        if self.target_path:
            table_target = DeltaTable.forPath(spark, self.target_path)
        else:
            table_target = DeltaTable.forName(spark, self.target_name)

        if self.scd_type == 1:
            if self.delete_where:
                delete_condition = F.coalesce(
                    F.expr(self.source_delete_where), F.lit(False)
                )
                not_delete_condition = ~delete_condition

            # Define merge
            conditions = [f"source.{c} = target.{c}" for c in self.primary_keys]
            merge = table_target.alias("target").merge(
                source.alias("source"),
                condition=" AND ".join(conditions),
            )

            # Update
            _set = {f"target.{c}": f"source.{c}" for c in self.update_columns}
            if self.ignore_null_updates:
                _set = {
                    f"target.{c}": F.coalesce(
                        F.col(f"source.{c}"), F.col(f"target.{c}")
                    ).alias(c)
                    for c in self.update_columns
                }

            condition = None
            if self.delete_where:
                condition = not_delete_condition
            if self.order_by:
                _condition = F.expr(f"source.{self.order_by} > target.{self.order_by}")
                if condition is None:
                    condition = _condition
                else:
                    condition = condition & _condition

            merge = merge.whenMatchedUpdate(set=_set, condition=condition)

            # Insert
            condition = None
            if self.delete_where:
                condition = not_delete_condition
            merge = merge.whenNotMatchedInsert(
                values={f"target.{c}": f"source.{c}" for c in self.write_columns},
                condition=condition,
            )

            # Delete
            if self.delete_where:
                merge = merge.whenMatchedDelete(condition=delete_condition)

            logger.info("Executing merge...")
            merge.execute()

        elif self.scd_type == 2:
            if self.delete_where:
                delete_condition = F.coalesce(F.expr(self.delete_where), F.lit(False))
                not_delete_condition = ~delete_condition

            # Only select rows that have been updated
            if self.target_path:
                target = spark.read.format("delta").load(self.target_path)
            else:
                target = spark.read.table(self.target_name)

            _source = source
            _target = target
            _on = [self.hash_cols] + self.primary_keys
            if self.delete_where:
                _source = source.withColumn("__to_delete", delete_condition)
                _target = target.withColumn("__to_delete", F.lit(False))
                _on += ["__to_delete"]

            upsert_or_delete = _source.join(
                other=_target,
                on=_on,
                how="leftanti",
            )

            # Merge
            conditions = []
            condition = F.expr(f"target.{self.end_at} IS NULL")
            for c in self.primary_keys:
                condition = condition & F.expr(f"source.{c} = target.{c}")
            merge = table_target.alias("target").merge(
                upsert_or_delete.filter(F.col(self.end_at).isNull()).alias("source"),
                condition=condition,
            )

            # Expire the current record
            _set = {f"target.{self.end_at}": f"source.{self.index_fist}"}
            merge = merge.whenMatchedUpdate(set=_set)
            #
            # # TODO: Review if required
            # if not self.delete_where:
            #     _set = {f"target.{self.end_at}": f"source.{self.order_by}"}
            #     merge = merge.whenMatchedUpdate(set=_set)
            # else:
            #     where = F.expr(self.source_delete_where)
            #     # deleting
            #     # _set = {f"target.{self.end_at}": "NULL"}
            #     _set = {f"target.{self.end_at}": f"source.{self.order_by}"}
            #     merge = merge.whenMatchedUpdate(set=_set, condition=where)
            #
            #     # updating
            #     _set = {f"target.{self.end_at}": f"source.{self.order_by}"}
            #     merge = merge.whenMatchedUpdate(set=_set, condition=~where)

            logger.info("Executing merge...")
            merge.execute()

            # Append rows
            upsert = upsert_or_delete
            if self.delete_where:
                upsert = upsert.filter(not_delete_condition)
            writer = (
                upsert.select(self.write_columns).write.mode("APPEND").format("DELTA")
            )
            logger.info("Appending new rows...")
            if self.target_path:
                writer.save(self.target_path)
            else:
                writer.saveAsTable(self.target_name)

        else:
            raise ValueError(f"SCD Type {self.scd_type} is not supported.")

    def execute(self, source: AnyFrame):
        """
        Merge source into target delta from sink

        Parameters
        ----------
        source:
            Source DataFrame to merge into target (sink).
        """

        dataframe_backend = DataFrameBackends.from_nw_implementation(
            source.implementation
        )
        if dataframe_backend not in SUPPORTED_BACKENDS:
            raise NotImplementedError(
                f"DataFrame provided is of {dataframe_backend} backend, which is not currently implemented for merge operations."
            )

        source = source.to_native()

        from delta.tables import DeltaTable

        self._source_schema = source.schema
        spark = source.sparkSession

        if self.target_path:
            if not DeltaTable.isDeltaTable(spark, self.target_path):
                self._init_target(source)
        else:
            try:
                spark.catalog.getTable(self.target_name)
            except Exception:
                self._init_target(source)

        if source.isStreaming:
            if self.sink is None:
                raise ValueError("Sink value required to fetch checkpoint location.")

            if self.sink and self.sink.checkpoint_path is None:
                raise ValueError(
                    f"Checkpoint location not specified for sink '{self.sink}'"
                )

            query = (
                source.writeStream.foreachBatch(
                    lambda batch_df, batch_id: self._execute(source=batch_df)
                )
                .trigger(availableNow=True)
                .options(
                    checkpointLocation=self.sink.checkpoint_path,
                )
                .start()
            )
            query.awaitTermination()

        else:
            self._execute(source=source)



from ._version import VERSION

__version__ = VERSION

# Import first
from ._settings import settings
from ._useragent import set_databricks_sdk_upstream

set_databricks_sdk_upstream()

import laktory._parsers
import laktory.api
import laktory.enums
import laktory.models
import laktory.narwhals_ext
import laktory.typing
import laktory.yaml

from ._cache import cache_dir
from ._logger import get_logger
from ._settings import Settings
from .sqlparser import SQLParser
from .version import show_version_info


def register_spark_session(spark=None):
    """Register a Spark session"""
    import sys

    _laktory = sys.modules[__name__]
    if spark is None:
        from pyspark.sql import SparkSession

        spark = (
            SparkSession.builder.appName("laktory")
            # TODO: Check if we can install on a need-basis
            .config(
                "spark.jars.packages",
                "org.apache.spark:spark-avro_2.12:3.5.5,io.delta:delta-spark_2.12:3.3.0",
            )  # com.databricks:spark-xml_2.12:0.17.0
            .config(
                "spark.sql.catalog.spark_catalog",
                "org.apache.spark.sql.delta.catalog.DeltaCatalog",
            )
            .getOrCreate()
        )
    _laktory._spark = spark


def get_spark_session():
    import sys

    _laktory = sys.modules[__name__]
    if not hasattr(_laktory, "_spark"):
        register_spark_session()
    return _laktory._spark


def is_dlt_execute() -> bool:
    from pyspark.errors import AnalysisException

    spark = get_spark_session()
    try:
        v = spark.conf.get("pipelines.dbrVersion", None)
    except AnalysisException:
        # Default value is not supported on serverless
        v = None

    return v is not None


def print_version():
    print(show_version_info())



from pathlib import Path
from typing import Any
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field
from pydantic import model_validator

from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class AlertConditionThresholdValue(BaseModel):
    bool_value: bool = Field(
        None,
        description="boolean value (`true` or `false`) to compare against boolean results.",
    )
    double_value: float = Field(
        None, description="double value to compare against integer and double results."
    )
    string_value: str = Field(
        None, description="string value to compare against string results."
    )


class AlertConditionThreshold(BaseModel):
    value: AlertConditionThresholdValue = Field(
        ...,
        description="Actual value used in comparison (one of the attributes is required)",
    )


class AlertConditionOperandColumn(BaseModel):
    name: str = Field(..., description="Name of the column")


class AlertConditionOperand(BaseModel):
    column: AlertConditionOperandColumn = Field(
        ...,
        description="""
    Block describing the column from the query result to use for comparison in alert evaluation
    """,
    )


class AlertCondition(BaseModel):
    op: str = Field(
        ...,
        description="""
    Operator used to compare in alert evaluation. (Enum: `GREATER_THAN`, `GREATER_THAN_OR_EQUAL`, `LESS_THAN`, 
    `LESS_THAN_OR_EQUAL`, `EQUAL`, `NOT_EQUAL`, `IS_NULL`)
    """,
    )
    operand: AlertConditionOperand = Field(
        ...,
        description="Name of the column from the query result to use for comparison in alert evaluation.",
    )
    empty_result_state: str = Field(
        None,
        description="Alert state if the result is empty (`UNKNOWN`, `OK`, `TRIGGERED`)",
    )
    threshold: AlertConditionThreshold = Field(
        ..., description="Threshold value used for comparison in alert evaluation:"
    )


class Alert(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks SQL Query

    Examples
    --------
    ```py
    from laktory import models

    alert = models.resources.databricks.Alert(
        display_name="My Alert",
        parent_path="/alerts",
        query_id="1",
        condition={
            "op": "GREATER_THAN",
            "operand": {"column": {"name": "value"}},
            "threshold": {"value": {"double_value": 42.0}},
        },
    )
    ```
    """

    access_controls: list[AccessControl] = Field(
        [], description="SQL Alert access controls"
    )
    condition: AlertCondition = Field(
        ..., description="Trigger conditions of the alert."
    )
    custom_body: str = Field(
        None,
        description="""
    Custom body of alert notification, if it exists. See [Alerts API reference](https://docs.databricks.com/sql/user/alerts/index.html)
        for custom templating instructions.
    """,
    )
    custom_subject: str = Field(
        None,
        description="""
    Custom subject of alert notification, if it exists. This includes email subject, Slack notification header, etc.
    See [Alerts API reference](https://docs.databricks.com/sql/user/alerts/index.html) for custom templating
    instructions.
    """,
    )
    display_name: str = Field(..., description="Name of the alert.")
    name_prefix: str = Field(None, description="Prefix added to the alert display name")
    name_suffix: str = Field(None, description="Suffix added to the alert display name")
    notify_on_ok: bool = Field(
        None,
        description="Whether to notify alert subscribers when alert returns back to normal.",
    )
    owner_user_name: str = Field(None, description="Alert owner's username.")
    parent_path_: str | None = Field(
        None,
        description="""
        The path to a workspace folder (inside laktory root) containing the alert. If changed, the query will be
        recreated.
        """,
        validation_alias=AliasChoices("parent_path", "dirpath", "parent_path_"),
        exclude=True,
    )
    query_id: str = Field(
        None,
        description="ID of the query evaluated by the alert. Mutually exclusive with `query`.",
    )
    seconds_to_retrigger: int = Field(
        None,
        description="""
    Number of seconds an alert must wait after being triggered to rearm itself. After rearming, it can be triggered 
    again. If 0 or not specified, the alert will not be triggered again.
    """,
    )

    @computed_field(description="parent_path")
    @property
    def parent_path(self) -> str:
        if self.parent_path_ is None:
            self.parent_path_ = ""
        if self.parent_path_.startswith("/"):
            self.parent_path_ = self.parent_path_[1:]

        parent_path = Path(settings.workspace_laktory_root) / self.parent_path_
        return parent_path.as_posix()

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.display_name = self.name_prefix + self.display_name
                self.name_prefix = ""
            if self.name_suffix:
                self.display_name = self.display_name + self.name_suffix
                self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.display_name.replace(" ", "_")

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    sql_alert_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Alert"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath", "name_prefix", "name_suffix"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_alert"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Literal

from pydantic import Field

from laktory.models.pipeline.orchestrators.pipelineconfigworkspacefile import (
    PipelineConfigWorkspaceFile,
)
from laktory.models.pipelinechild import PipelineChild
from laktory.models.resources.databricks.cluster import ClusterLibrary
from laktory.models.resources.databricks.cluster import ClusterLibraryPypi
from laktory.models.resources.databricks.job import Job
from laktory.models.resources.databricks.job import JobEnvironment
from laktory.models.resources.databricks.job import JobEnvironmentSpec
from laktory.models.resources.databricks.job import JobParameter
from laktory.models.resources.databricks.job import JobTask
from laktory.models.resources.databricks.job import JobTaskPythonWheelTask
from laktory.models.resources.pulumiresource import PulumiResource

ENV_KEY = "laktory"


class DatabricksJobOrchestrator(Job, PipelineChild):
    """
    Databricks job used as an orchestrator to execute a Laktory pipeline.

    Job orchestrator supports incremental workloads with Spark Structured
    Streaming, but it does not support continuous processing.

    References
    ----------
    * [Data Pipeline](https://www.laktory.ai/concepts/pipeline/)
    * [Databricks Job](https://docs.databricks.com/en/workflows/jobs/create-run-jobs.html)
    """

    type: Literal["DATABRICKS_JOB"] = Field(
        "DATABRICKS_JOB", description="Type of orchestrator"
    )
    config_file: PipelineConfigWorkspaceFile = Field(
        PipelineConfigWorkspaceFile(),
        description="Pipeline configuration (json) file deployed to the workspace and used by the job to read and execute the pipeline.",
    )
    node_max_retries: int = Field(
        None,
        description="An optional maximum number of times to retry an unsuccessful run for each node.",
    )

    # ----------------------------------------------------------------------- #
    # Update Job                                                              #
    # ----------------------------------------------------------------------- #

    def _set_task_compute(self, task, serverless, _requirements):
        if serverless:
            task.environment_key = ENV_KEY
        else:
            libraries = []
            for r in _requirements:
                is_var = "${vars." in r
                if r.endswith(".whl") or (is_var and "wheel" in r or "whl" in r):
                    l = ClusterLibrary(whl=r)
                else:
                    l = ClusterLibrary(pypi=ClusterLibraryPypi(package=r))
                libraries += [l]

            task.job_cluster_key = "node-cluster"
            task.libraries = libraries
        return task

    def update_from_parent(self):
        serverless = True
        for c in self.job_clusters:
            if c.job_cluster_key == "node-cluster":
                serverless = False
        if len(self.job_clusters) > 0 and serverless:
            raise ValueError(
                "To use DATABRICKS_JOB orchestrator, a cluster named `node-cluster` must be defined in the databricks_job attribute."
            )

        pl = self.parent_pipeline

        self.tasks = []

        def _get_depends_on(node, pl):
            depends_on = []
            for edge in pl.dag.in_edges(node.name):
                _node = pl.nodes_dict[edge[0]]
                if _node.has_sinks:
                    depends_on += [{"task_key": "node-" + edge[0]}]
                else:
                    depends_on += _get_depends_on(_node, pl)
            return depends_on

        # Requirements and path
        _requirements = self.inject_vars_into_dump({"deps": pl._dependencies})["deps"]
        _path = (
            "/Workspace"
            + self.inject_vars_into_dump({"path": self.config_file.path})["path"]
        )

        # Environment
        env_found = False
        envs = self.environments
        if envs is None:
            envs = []
        for env in envs:
            if env.environment_key == ENV_KEY:
                env_found = True
                break
        if not env_found:
            envs += [
                JobEnvironment(
                    environment_key=ENV_KEY,
                    spec=JobEnvironmentSpec(
                        client="3",
                        dependencies=_requirements,
                    ),
                )
            ]
            self.environments = envs

        # Sorting Node Names to prevent job update trigger with Pulumi
        node_names = [node.name for node in pl.nodes]
        node_names.sort()
        for node_name in node_names:
            node = pl.nodes_dict[node_name]

            if not node.has_sinks:
                continue

            depends_on = _get_depends_on(node, pl=pl)

            task = JobTask(
                task_key="node-" + node.name,
                python_wheel_task=JobTaskPythonWheelTask(
                    entry_point="models.pipeline._execute",
                    package_name="laktory",
                    named_parameters={
                        "filepath": _path,
                        "node_name": node.name,
                    },
                ),
                depends_ons=depends_on,
            )
            task = self._set_task_compute(task, serverless, _requirements)

            if self.node_max_retries:
                task.max_retries = self.node_max_retries

            self.tasks += [task]

        if pl.databricks_quality_monitor_enabled:
            task = JobTask(
                task_key="post-execute",
                python_wheel_task=JobTaskPythonWheelTask(
                    entry_point="models.pipeline._post_execute",
                    package_name="laktory",
                    named_parameters={
                        "filepaths": _path,
                        "tables_metadata": "false",  # this has been already update in the pl.execute()
                        "quality_monitors": "true",  # this has been already update in the pl.execute()
                    },
                ),
                depends_ons=[{"task_key": t.task_key} for t in self.tasks],
            )
            task = self._set_task_compute(task, serverless, _requirements)
            self.tasks += [task]

        self.sort_tasks(self.tasks)

        # Update job parameters
        self.parameters = [
            JobParameter(name="full_refresh", default="false"),
        ]

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def children_names(self):
        return ["config_file"]

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        return "job"

    @property
    def pulumi_excludes(self) -> list[str] | dict[str, bool]:
        return super().pulumi_excludes + [
            "config_file",
            "node_max_retries",
            "type",
            "dataframe_backend",
            "dataframe_api",
        ]

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - configuration workspace file
        - configuration workspace file permissions
        """

        resources = super().additional_core_resources
        resources += [self.config_file]

        return resources



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class NotificationDestinationConfigEmail(BaseModel):
    """Notification Destination Config Email"""

    addresses: list[str] = Field(
        ..., description="The list of email addresses to send notifications to."
    )


class NotificationDestinationConfigGenericWebhook(BaseModel):
    """Notification Destination Config Generic Webhook"""

    password: str = Field(None, description="The password for basic authentication.")
    password_set: str = Field(None, description="")
    url: str = Field(None, description="The Generic Webhook URL.")
    url_set: str = Field(None, description="")
    username: str = Field(None, description="The username for basic authentication.")
    username_set: str = Field(None, description="")


class NotificationDestinationConfigMicrosoftTeams(BaseModel):
    """Notification Destination Config Microsoft Teams"""

    url: str = Field(..., description="The Microsoft Teams webhook URL")
    url_set: str = Field(None, description="")


class NotificationDestinationConfigPagerduty(BaseModel):
    """Notification Destination Config PagerDuty"""

    integration_key: str = Field(..., description="The PagerDuty integration key")
    integration_key_set: str = Field(None, description="")


class NotificationDestinationConfigSlack(BaseModel):
    """Notification Destination Config Slack"""

    url: str = Field(..., description="The Slack webhook URL")
    url_set: str = Field(None, description="")


class NotificationDestinationConfig(BaseModel):
    email: NotificationDestinationConfigEmail = Field(
        None, description="The email configuration of the Notification Destination."
    )
    generic_webhook: NotificationDestinationConfigGenericWebhook = Field(
        None,
        description="The Generic Webhook configuration of the Notification Destination.",
    )
    microsoft_teams: NotificationDestinationConfigMicrosoftTeams = Field(
        None,
        description="The Microsoft Teams configuration of the Notification Destination.",
    )
    pagerduty: NotificationDestinationConfigPagerduty = Field(
        None, description="The PagerDuty configuration of the Notification Destination."
    )
    slack: NotificationDestinationConfigSlack = Field(
        None, description="The Slack configuration of the Notification Destination."
    )

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "microsoft_teams": "microsoft_teams",
        }


class NotificationDestination(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks notification destination

    Examples
    --------
    ```py
    import laktory as lk

    nd_slack = lk.models.resources.databricks.NotificationDestination(
        display_name="slack",
        config={"slack": {"url": "slack.webhook.com"}},
    )
    ```
    """

    display_name: str = Field(
        ..., description="The display name of the Notification Destination."
    )
    config: NotificationDestinationConfig = Field(
        None, description="The configuration of the Notification Destination"
    )
    # destination_type: str = Field(None, description="The type of Notification Destination.")

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:NotificationDestination"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_notification_destination"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Literal

from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.enums import DataFrameBackends
from laktory.models.basechild import BaseChild

logger = get_logger(__name__)


class PipelineChild(BaseChild):
    """
    Pipeline Child Class
    """

    # model_config = ConfigDict(validate_assignment=True)
    dataframe_backend_: DataFrameBackends = Field(
        None,
        description="Type of DataFrame backend",
        validation_alias=AliasChoices("dataframe_backend", "dataframe_backend_"),
        exclude=True,
    )
    dataframe_api_: Literal["NARWHALS", "NATIVE"] = Field(
        None,
        description="""
        DataFrame API to use in DataFrame Transformer nodes. Either 'NATIVE' (backend-specific) or 'NARWHALS' 
        (backend-agnostic).
        """,
        validation_alias=AliasChoices("dataframe_api", "dataframe_api_"),
        exclude=True,
    )

    @computed_field(description="dataframe_backend")
    @property
    def dataframe_backend(self) -> DataFrameBackends:
        backend = self.dataframe_backend_

        # Direct value
        if backend is not None:
            if not isinstance(backend, DataFrameBackends):
                try:
                    backend = DataFrameBackends(backend)
                except ValueError:
                    # TODO: Review why this might occur
                    pass

            return backend

        # Value from parent
        parent = self._parent
        if parent is not None:
            return parent.dataframe_backend

        # Value from settings
        return DataFrameBackends(settings.dataframe_backend.upper())

    @computed_field(description="dataframe_api")
    @property
    def dataframe_api(self) -> str:
        # Direct value
        if self.dataframe_api_:
            return self.dataframe_api_

        # Value from parent
        parent = self._parent
        if parent is not None:
            return parent.dataframe_api

        # Value from settings
        return settings.dataframe_api.upper()

    @property
    def parent_pipeline(self):
        from laktory.models.pipeline.pipeline import Pipeline

        def _get_pl(o):
            parent = getattr(o, "_parent", None)
            if parent is None:
                return None

            if isinstance(parent, Pipeline):
                return parent

            return _get_pl(parent)

        return _get_pl(self)

    @property
    def parent_pipeline_node(self):
        from laktory.models.pipeline.pipelinenode import PipelineNode

        def _get_pl(o):
            parent = getattr(o, "_parent", None)
            if parent is None:
                return None

            if isinstance(parent, PipelineNode):
                return parent

            return _get_pl(parent)

        return _get_pl(self)



from typing import Literal

from narwhals import LazyFrame
from pydantic import Field

from laktory._logger import get_logger
from laktory.models.datasources.tabledatasource import TableDataSource

logger = get_logger(__name__)


# class Connection(BaseModel):
#     workspace_url: str
#     bearer_token: str = "auto"
#     require_https: bool = True


class UnityCatalogDataSource(TableDataSource):
    """
    Data source using a Unity Catalog data table.

    Examples
    ---------
    ```python
    from laktory import models

    source = models.UnityCatalogDataSource(
        catalog_name="dev",
        schema_name="finance",
        table_name="brz_stock_prices",
        selects=["symbol", "open", "close"],
        filter="symbol='AAPL'",
        as_stream=True,
    )
    # df = source.read()
    ```
    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/sourcessinks/)
    """

    # connection: Connection = None
    type: Literal["UNITY_CATALOG"] = Field(
        "UNITY_CATALOG", frozen=True, description="Source Type"
    )

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_polars(self) -> LazyFrame:
        # TODO
        # https://docs.pola.rs/api/python/stable/reference/catalog/api/polars.Catalog.scan_table.html#polars.Catalog.scan_table
        raise NotImplementedError(
            "Unity Catalog data source is not yet implemented for Polars"
        )



import os

from .get_dfs import StreamingSource
from .get_dfs import get_df0
from .get_dfs import get_df1
from .paths import Paths


def skip_test(required, extras=None):
    import pytest

    if extras is not None:
        for e in extras:
            required += [e]

    missing = []
    for k in required:
        if not os.getenv(k):
            missing += [k]
    if missing:
        pytest.skip(f"{missing} not available.")


def skip_pulumi_preview(extras=None):
    skip_test(
        required=[
            "PULUMI_ACCESS_TOKEN",
            "DATABRICKS_HOST",
            "DATABRICKS_TOKEN",
        ],
        extras=extras,
    )


def skip_terraform_plan(extras=None):
    skip_test(
        required=[
            "DATABRICKS_HOST",
            "DATABRICKS_TOKEN",
        ],
        extras=extras,
    )



import os
from pathlib import Path
from typing import Any
from typing import Literal

import narwhals as nw
from pydantic import AliasChoices
from pydantic import ConfigDict
from pydantic import Field
from pydantic import computed_field
from pydantic import field_serializer
from pydantic import field_validator
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.dataframe.dataframeschema import DataFrameSchema
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.models.readerwritermethod import ReaderWriterMethod

logger = get_logger(__name__)

SUPPORTED_FORMATS = {
    DataFrameBackends.PYSPARK: [
        "AVRO",
        "BINARYFILE",
        "CSV",
        "DELTA",
        "JSON",
        "JSONL",
        "NDJSON",  # SAME AS JSONL
        "ORC",
        "PARQUET",
        "TEXT",
        "XML",
    ],
    DataFrameBackends.POLARS: [
        "AVRO",
        "CSV",
        "DELTA",
        "EXCEL",
        "IPC",
        # "ICEBERG", # TODO
        "JSON",
        "JSONL",
        "NDJSON",  # SAME AS JSONL
        "PARQUET",
        "PYARROW",
    ],
}

ALL_SUPPORTED_FORMATS = tuple(sorted(set().union(*SUPPORTED_FORMATS.values())))


class FileDataSource(BaseDataSource):
    """
    Data source using disk files, such data events (json/csv) or full dataframes.

    Examples
    ---------
    ```python
    from laktory import models

    source = models.FileDataSource(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="JSON",
        dataframe_backend="POLARS",
    )
    # df = source.read()

    # With Explicit Schema
    source = models.FileDataSource(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="JSON",
        dataframe_backend="PYSPARK",
        schema={
            "columns": {
                "symbol": "String",
                "open": "Float64",
                "close": "Float64",
            }
        },
    )
    # df = source.read()
    ```

    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/sourcessinks/)
    * [autoloader schema inference](https://docs.databricks.com/aws/en/ingestion/cloud-object-storage/auto-loader/schema)
    """

    model_config = ConfigDict(populate_by_name=True)
    format: Literal.__getitem__(ALL_SUPPORTED_FORMATS) = Field(
        ..., description="Format of the data files."
    )
    has_header: bool = Field(
        True,
        description="Indicate if the first row of the dataset is a header or not. Only applicable to 'CSV' format.",
    )
    infer_schema: bool = Field(
        False,
        description="""
        When `True`, the schema is inferred from the data. When `False`, the schema is not inferred and will be string 
        if not specified in schema_definition. Only applicable to some format like CSV and JSON.
        """,
    )
    path: str = Field(
        ...,
        description="File path on a local disk, remote storage or Databricks volume.",
    )
    reader_kwargs: dict[str, Any] = Field(
        {},
        description="""
        Keyword arguments passed directly to dataframe backend reader. Passed to `.options()` method when using PySpark.
        """,
    )
    schema_definition: DataFrameSchema = Field(
        None,
        validation_alias="schema",
        description="""
        Target schema specified as a list of columns, as a dict or a json serialization. Only used when
        reading data from non-strongly typed files such as JSON or csv files.
        """,
    )
    schema_location_: str | Path = Field(
        None,
        description="Path for schema inference when reading data as a stream. If `None`, parent directory of `path` is used.",
        validation_alias=AliasChoices("schema_location", "schema_location_"),
        exclude=True,
    )
    type: Literal["FILE"] = Field("FILE", frozen=True, description="Source Type")
    reader_methods: list[ReaderWriterMethod] = Field(
        [], description="DataFrame backend reader methods."
    )
    # schema_overrides: DataFrameSchema = Field(None, validation_alias="schema")

    @field_validator("path", mode="before")
    @classmethod
    def posixpath_to_string(cls, value: Any) -> Any:
        if isinstance(value, Path):
            value = str(value)
        return value

    @model_validator(mode="after")
    def validate_format(self) -> Any:
        if self.format not in SUPPORTED_FORMATS[self.dataframe_backend]:
            raise ValueError(
                f"'{self.format}' format is not supported with {self.dataframe_backend}. Use one of {SUPPORTED_FORMATS[self.dataframe_backend]}"
            )
        return self

    @model_validator(mode="after")
    def validate_options(self) -> Any:
        for k in [
            "has_header",
            "infer_schema",
            "schema_definition",
            "schema_location",
        ]:  # "schema_overrides",
            if k in self.model_fields_set:
                if not self._is_applicable(k):
                    raise ValueError(
                        f"Attribute `{k}` is not supported for the provider source format / backend / stream"
                    )
        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self.path)

    @computed_field(description="schema_location")
    @property
    def schema_location(self) -> Path:
        if self.schema_location_:
            return Path(self.schema_location_)

        return Path(os.path.dirname(self.path))

    @field_serializer("schema_location", when_used="json")
    def serialize_path(self, value: Path) -> str:
        return value.as_posix()

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _is_applicable(self, key):
        if key == "has_header":
            if self.format == "CSV":
                return True
            return False

        if key == "infer_schema":
            if self.format in ["CSV", "XML"]:
                return True

            if self.as_stream:
                # https://docs.databricks.com/aws/en/ingestion/cloud-object-storage/auto-loader/schema
                if self.format in [
                    "AVRO",
                    "CSV",
                    "JSON",
                    "JSONL",
                    "NDJSON",
                    "PARQUET",
                    "XML",
                ]:
                    return True

            return False

        if key == "schema_definition":
            if self.dataframe_backend == DataFrameBackends.PYSPARK:
                if self.format in [
                    "AVRO",
                    "CSV",
                    "DELTA",
                    "JSON",
                    "JSONL",
                    "NDJSON",
                    "ORC",
                    "PARQUET",
                ]:
                    return True

            if self.dataframe_backend == DataFrameBackends.POLARS:
                if self.format in ["CSV", "JSON", "JSONL", "NDJSON", "PARQUET"]:
                    return True

            return False

        if key == "schema_location":
            return self.is_cloud_files

        return False

    @property
    def is_cloud_files(self):
        return self.as_stream and self.format not in ["DELTA"]

    def _get_spark_kwargs(self):
        fmt = self.format.lower()

        # Build kwargs
        kwargs = {}

        if fmt in ["jsonl", "ndjson"]:
            kwargs["multiline"] = False
            fmt = "json"
        elif fmt in ["json"]:
            kwargs["multiline"] = True

        if self._is_applicable("has_header"):
            kwargs["header"] = self.has_header

        if self._is_applicable("infer_schema"):
            if self.as_stream:
                kwargs["cloudFiles.inferColumnTypes"] = self.infer_schema
            else:
                kwargs["inferSchema"] = self.infer_schema

        if self.as_stream:
            # Cloud Files Formats
            if self.is_cloud_files:
                kwargs["cloudFiles.format"] = fmt
                kwargs["recursiveFileLookup"] = True
                kwargs["cloudFiles.schemaLocation"] = self.schema_location.as_posix()
                fmt = "cloudFiles"

            # Native Streaming Formats
            else:
                pass

        for k, v in self.reader_kwargs.items():
            kwargs[k] = v

        return kwargs, fmt

    def _get_spark_reader_methods(self):
        methods = []

        options, fmt = self._get_spark_kwargs()

        if self.schema_definition:
            methods += [
                ReaderWriterMethod(
                    name="schema", args=[self.schema_definition.to_spark()]
                )
            ]

        methods += [ReaderWriterMethod(name="format", args=[fmt])]

        if options:
            methods += [ReaderWriterMethod(name="options", kwargs=options)]

        for m in self.reader_methods:
            methods += [m]

        return methods

    def _read_spark(self) -> nw.LazyFrame:
        from laktory import get_spark_session

        spark = get_spark_session()

        # Create reader
        if self.as_stream:
            mode = "stream"
            reader = spark.readStream
        else:
            mode = "static"
            reader = spark.read

        # Build methods
        methods = self._get_spark_reader_methods()

        # Apply methods
        for m in methods:
            reader = getattr(reader, m.name)(*m.args, **m.kwargs)

        # Load
        logger.info(
            f"Reading {self.path} as {mode} read.{'.'.join([m.as_string for m in methods])}"
        )
        df = reader.load(self.path)

        return nw.from_native(df)

    def _get_polars_kwargs(self):
        fmt = self.format.lower()

        kwargs = {}

        if self._is_applicable("has_header") and self.has_header is not None:
            kwargs["has_header"] = self.has_header

        if self._is_applicable("infer_schema") and self.infer_schema is not None:
            kwargs["infer_schema"] = self.infer_schema

        if (
            self._is_applicable("schema_definition")
            and self.schema_definition is not None
        ):
            kwargs["schema"] = self.schema_definition.to_polars()

        for k, v in self.reader_kwargs.items():
            kwargs[k] = v

        return kwargs, fmt

    def _read_polars(self) -> nw.LazyFrame:
        import polars as pl

        kwargs, fmt = self._get_polars_kwargs()

        logger.info(f"Reading {self.path} with format '{fmt}' and {kwargs}")

        if fmt == "avro":
            df = pl.read_avro(self.path, **kwargs).lazy()

        elif fmt == "csv":
            df = pl.scan_csv(self.path, **kwargs)

        elif fmt == "delta":
            df = pl.scan_delta(self.path, **kwargs)

        elif fmt == "excel":
            df = pl.read_excel(self.path, **kwargs).lazy()

        elif fmt == "ipc":
            df = pl.scan_ipc(self.path, **kwargs)

        elif fmt == "iceberg":
            df = pl.scan_iceberg(self.path, **kwargs)

        elif fmt == "json":
            df = pl.read_json(self.path, **kwargs).lazy()

        elif fmt in ["ndjson", "jsonl"]:
            df = pl.scan_ndjson(self.path, **kwargs)

        elif fmt == "parquet":
            df = pl.scan_parquet(self.path, **kwargs)

        elif fmt == "pyarrow":
            import pyarrow.dataset as ds

            dset = ds.dataset(self.path, format="parquet")
            return pl.scan_pyarrow_dataset(dset, **kwargs)

        else:
            raise ValueError(f"Format {fmt} is not supported.")

        return nw.from_native(df)



import os
import shutil
from pathlib import Path
from typing import Any
from typing import Literal

from pydantic import Field
from pydantic import field_validator

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.datasinks.basedatasink import POLARS_DELTA_MODES
from laktory.models.datasinks.basedatasink import BaseDataSink
from laktory.models.datasources.filedatasource import FileDataSource

SUPPORTED_FORMATS = {
    DataFrameBackends.PYSPARK: [
        "AVRO",
        "BINARYFILE",
        "CSV",
        "DELTA",
        "JSON",
        "JSONL",
        "NDJSON",  # SAME AS JSONL
        "ORC",
        "PARQUET",
        "TEXT",
        "XML",
    ],
    DataFrameBackends.POLARS: [
        "AVRO",
        "CSV",
        "DELTA",
        "EXCEL",
        "IPC",
        # "ICEBERG", # TODO
        "JSON",
        "JSONL",
        "NDJSON",  # SAME AS JSONL
        "PARQUET",
        "PYARROW",
    ],
}

ALL_SUPPORTED_FORMATS = tuple(sorted(set().union(*SUPPORTED_FORMATS.values())))

logger = get_logger(__name__)


class FileDataSink(BaseDataSink):
    """
    Data sink writing to disk file(s) as csv, parquet or Delta Table.

    Examples
    ---------
    Write polars DataFrame as CSV
    ```python
    import polars as pl

    import laktory as lk

    df = pl.DataFrame({"x": [0, 1]})

    sink = lk.models.FileDataSink(
        path="./dataframe.csv", format="CSV", writer_kwargs={"separator": ";"}
    )
    sink.write(df)
    ```

    Write Spark Streaming DataFrame as Delta
    ```python tag:skip-run
    from laktory import models

    df = spark.readStream(...)  # skip

    sink = models.FileDataSink(
        path="./delta_table/",
        format="DELTA",
        mode="APPEND",
        checkpoint_path="./delta_table/checkpoint/",
    )
    sink.write(df)
    ```
    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/sourcessinks/)
    """

    format: Literal.__getitem__(ALL_SUPPORTED_FORMATS) = Field(
        ..., description="Format of the data files."
    )
    path: str = Field(
        ...,
        description="File path on a local disk, remote storage or Databricks volume.",
    )
    type: Literal["FILE"] = Field("FILE", frozen=True, description="Source Type")

    @field_validator("path", mode="before")
    @classmethod
    def posixpath_to_string(cls, path: Any) -> Any:
        if isinstance(path, Path):
            path = str(path)
        return path

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return self.path

    # ----------------------------------------------------------------------- #
    # Writers                                                                 #
    # ----------------------------------------------------------------------- #

    def _validate_format(self) -> None:
        if self.format not in SUPPORTED_FORMATS[self.dataframe_backend]:
            raise ValueError(
                f"'{self.format}' format is not supported with {self.dataframe_backend}. Use one of {SUPPORTED_FORMATS[self.dataframe_backend]}"
            )
        if self.mode == "MERGE" and self.format != "DELTA":
            raise ValueError("Only 'DELTA' format is supported with 'MERGE' mode.")

    def _validate_mode_polars(self, mode, df):
        if self.format == "DELTA":
            if mode not in POLARS_DELTA_MODES:
                raise ValueError(
                    f"Mode '{mode}' is not supported for Polars DataFrame with DELTA format. Set to {POLARS_DELTA_MODES}"
                )
        else:
            if mode:
                raise ValueError(
                    f"Mode '{mode}' is not supported for Polars DataFrame. Set to `None`"
                )

    def _write_spark(self, df, mode, full_refresh=False) -> None:
        df = df.to_native()

        # Format
        methods = self._get_spark_writer_methods(mode=mode, is_streaming=df.isStreaming)

        if df.isStreaming:
            logger.info(
                f"Writing df to {self.path} with writeStream.{'.'.join([m.as_string for m in methods])}"
            )

            writer = df.writeStream
            for m in methods:
                writer = getattr(writer, m.name)(*m.args, **m.kwargs)

            query = writer.start(self.path)
            query.awaitTermination()

        else:
            logger.info(
                f"Writing df to {self.path} with write.{'.'.join([m.as_string for m in methods])}"
            )
            writer = df.write
            for m in methods:
                writer = getattr(writer, m.name)(*m.args, **m.kwargs)

            writer.save(self.path)

    def _write_polars(self, df, mode, full_refresh=False) -> None:
        import polars as pl

        kwargs, fmt = self._get_polars_kwargs(mode=mode)

        df = df.to_native()

        is_streaming = False

        _mode = "streaming" if is_streaming else "static"
        logger.info(
            f"Writing {_mode} df to {self.path} with format '{self.format}' and {kwargs}"
        )

        if isinstance(df, pl.LazyFrame):
            df = df.collect()

        if self.format.lower() == "avro":
            df.write_avro(self.path, **self.writer_kwargs)
        if self.format.lower() == "csv":
            df.write_csv(self.path, **self.writer_kwargs)
        elif self.format.lower() == "delta":
            df.write_delta(self.path, mode=mode, **self.writer_kwargs)
        elif self.format.lower() == "excel":
            df.write_excel(self.path, **self.writer_kwargs)
        elif self.format.lower() == "ipc":
            df.write_ipc(self.path, **self.writer_kwargs)
        elif self.format.lower() == "json":
            df.write_json(self.path, **self.writer_kwargs)
        elif self.format.lower() in ["ndjson", "jsonl"]:
            df.write_ndjson(self.path, **self.writer_kwargs)
        elif self.format.lower() == "parquet":
            df.write_parquet(self.path, **self.writer_kwargs)
        elif self.format.lower() == "pyarrow":
            import pyarrow.dataset as ds

            kwargs = {"format": "parquet", "partitioning": None}
            for k, v in self.writer_kwargs.items():
                kwargs[k] = v
            ds.write_dataset(data=df.to_arrow(), base_dir=self.path, **kwargs)

    # ----------------------------------------------------------------------- #
    # Purge                                                                   #
    # ----------------------------------------------------------------------- #

    def purge(self):
        """
        Delete sink data and checkpoints
        """
        # Remove Data
        if os.path.exists(self.path):
            is_dir = os.path.isdir(self.path)
            if is_dir:
                logger.info(f"Deleting data dir {self.path}")
                shutil.rmtree(self.path)
            else:
                logger.info(f"Deleting data file {self.path}")
                os.remove(self.path)

        # TODO: Add support for Databricks dbfs / workspace / Volume?

        # Remove Checkpoint
        self._purge_checkpoint()

    # ----------------------------------------------------------------------- #
    # Source                                                                  #
    # ----------------------------------------------------------------------- #

    def as_source(self, as_stream: bool = None) -> FileDataSource:
        """
        Generate a file data source with the same path as the sink.

        Parameters
        ----------
        as_stream:
            If `True`, sink will be read as stream.

        Returns
        -------
        :
            File Data Source
        """

        source = FileDataSource(
            path=self.path, format=self.format, dataframe_backend=self.dataframe_backend
        )

        if as_stream:
            source.as_stream = as_stream

        # if self.dataframe_backend:
        #     source.dataframe_backend = self.dataframe_backend
        source.parent = self.parent

        return source



import os
from pathlib import Path
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class DbfsFileLookup(ResourceLookup):
    path: str = Field(
        serialization_alias="id",
        description="Path on DBFS for the file from which to get content.",
    )
    limit_file_size: bool = Field(
        True, description="Do not load content for files larger than 4MB."
    )


class DbfsFile(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks DBFS File

    Examples
    --------
    ```py
    from laktory import models

    file = models.resources.databricks.DbfsFile(
        source="./data/stock_prices/prices.json",
    )
    print(file.path)
    # > /prices.json

    file = models.resources.databricks.DbfsFile(
        source="./data/stock_prices/prices.json",
    )
    print(file.path)
    # > /prices.json

    file = models.resources.databricks.DbfsFile(
        source="./data/stock_prices/prices.json",
        dirpath="stock_prices/",
    )
    print(file.path)
    # > /stock_prices/prices.json
    ```
    """

    access_controls: list[AccessControl] = Field(
        [], description="List of file access controls"
    )
    dirpath: str = Field(
        None,
        description="""
    Workspace directory inside rootpath in which the DBFS file is deployed. Used only if `path` is not specified.
    """,
    )
    lookup_existing: DbfsFileLookup = Field(
        None,
        exclude=True,
        description="Specifications for looking up existing resource. Other attributes will be ignored.",
    )
    path_: str = Field(
        None,
        description="DBFS filepath for the file. Overwrite `dirpath`.",
        validation_alias=AliasChoices("path", "path_"),
        exclude=True,
    )
    source: str = Field(..., description="Path to file on local filesystem.")

    @computed_field(description="path")
    @property
    def path(self) -> str | None:
        if self.path_:
            return self.path_

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        path = Path("/") / self.dirpath / self.filename
        return path.as_posix()

    @classmethod
    def lookup_defaults(cls) -> dict:
        return {"path": ""}

    @property
    def filename(self) -> str:
        """File filename"""
        return os.path.basename(self.source)

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.path

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    workspace_file_path=f"${{resources.{self.resource_name}.path}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:DbfsFile"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_dbfs_file"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import re
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import Union

import narwhals as nw
from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.basemodel import BaseModel
from laktory.models.pipelinechild import PipelineChild

if TYPE_CHECKING:
    import polars as pl
    import pyspark.sql.functions as F


logger = get_logger(__name__)

if TYPE_CHECKING:
    import polars as pl
    import pyspark.sql.functions as F


class DataFrameColumnExpr(BaseModel, PipelineChild):
    """
    DataFrame Column Expression defined with a string representation of DataFrame
    API expression (native or Narwhals) or a SQL statement.

    Examples
    --------
    Define serializable expressions in native DataFrame API.
    ```py
    import polars as pl

    import laktory as lk

    df = pl.DataFrame(
        {
            "x": [1, 2, 3],
        }
    )

    expr1 = lk.models.DataFrameColumnExpr(
        expr="pl.col('x')+pl.lit(1)",
        dataframe_backend="POLARS",
        dataframe_api="NATIVE",
    )

    expr2 = lk.models.DataFrameColumnExpr(
        expr="2*x + 1",
        type="SQL",
        dataframe_backend="POLARS",
        dataframe_api="NATIVE",
    )

    df = df.with_columns(
        y1=expr1.to_expr(),
        y2=expr2.to_expr(),
    )

    print(df)
    '''
    | x | y1 | y2 |
    |---|----|----|
    | 1 | 2  | 3  |
    | 2 | 3  | 5  |
    | 3 | 4  | 7  |
    '''
    ```

    Define serializable expressions in Narwhals DataFrame API.
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk

    df = nw.from_native(
        pl.DataFrame(
            {
                "x": [1, 2, 3],
            }
        )
    )

    expr1 = lk.models.DataFrameColumnExpr(
        expr="nw.col('x')+nw.lit(1)",
        dataframe_backend="POLARS",
        dataframe_api="NARWHALS",
    )

    expr2 = lk.models.DataFrameColumnExpr(
        expr="2*x + 1",
        type="SQL",
        dataframe_backend="POLARS",
        dataframe_api="NARWHALS",
    )

    df = df.with_columns(
        y1=expr1.to_expr(),
        y2=expr2.to_expr(),
    )

    print(df)
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    | | x | y1 | y2 |  |
    | |---|----|----|  |
    | | 1 | 2  | 3  |  |
    | | 2 | 3  | 5  |  |
    | | 3 | 4  | 7  |  |
    └──────────────────┘
    '''
    ```
    """

    expr: str = Field(..., description="Expression string representation")
    type: Literal["SQL", "DF"] = Field(
        None,
        description="Expression type: DF or SQL. If `None` is specified, type is guessed from provided expression.",
    )

    @model_validator(mode="after")
    def guess_type(self) -> Any:
        if self.type:
            return self

        expr_clean = self.expr.strip().replace("\n", " ")

        type = "SQL"
        if re.findall(r"\w+\.\w+\(", expr_clean):
            type = "DF"

        for k in [
            "lit(",
            "col(",
            "F.",
            "nw.",
            "pl.",
        ]:
            if k in expr_clean:
                type = "DF"
                break

        self.type = type

        return self

    # ----------------------------------------------------------------------- #
    # Expressions                                                             #
    # ----------------------------------------------------------------------- #

    def to_sql_expr(self) -> str:
        """Column expression expressed as a SQL Statement"""
        # -> pure SQL

        if self.type == "SQL":
            return self.expr
        else:
            # TODO: Use SQLFrame?
            raise ValueError("DataFrame expression can't be converted to SQL")

    def to_expr(self) -> Union[nw.Expr, "pl.Expr", "F.Column"]:
        """Column expression expressed as DataFrame API object"""

        _value = self.expr.replace("\n", " ")

        if self.dataframe_api == "NARWHALS":
            if self.type == "SQL":
                from laktory.narwhals_ext.functions import sql_expr

                expr = sql_expr(_value)
            else:
                # Imports required to evaluate expressions
                import narwhals as nw  # noqa: F401
                from narwhals import col  # noqa: F401
                from narwhals import lit  # noqa: F401

                expr = eval(_value)
        else:
            if self.dataframe_backend == DataFrameBackends.PYSPARK:
                if self.type == "SQL":
                    import pyspark.sql.functions as F

                    expr = F.expr(_value)
                else:
                    # Imports required to evaluate expressions
                    import pyspark.sql.functions as F  # noqa: F401
                    import pyspark.sql.types as T  # noqa: F401
                    from pyspark.sql.functions import col  # noqa: F401
                    from pyspark.sql.functions import lit  # noqa: F401

                    expr = eval(_value)

            elif self.dataframe_backend == DataFrameBackends.POLARS:
                if self.type == "SQL":
                    import polars as pl

                    expr = pl.sql_expr(_value)
                else:
                    # Imports required to evaluate expressions
                    import polars as pl  # noqa: F401
                    import polars.functions as F  # noqa: F401
                    from polars import col  # noqa: F401
                    from polars import lit  # noqa: F401

                    expr = eval(self.expr)

            else:
                raise ValueError(
                    f"`dataframe_backend` '{self.dataframe_backend}' is not supported."
                )

        return expr



import re


def has_column(self, col: str) -> bool:
    """
    Check if column `col` exists in `df`

    Parameters
    ----------
    col
        Column name

    Returns
    -------
    :
        Result

    Examples
    --------

    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk  # noqa: F401

    df = nw.from_native(
        pl.DataFrame(
            {
                "indexx": [1, 2, 3],
                "stock": [
                    {"symbol": "AAPL", "name": "Apple"},
                    {"symbol": "MSFT", "name": "Microsoft"},
                    {"symbol": "GOOGL", "name": "Google"},
                ],
                "prices": [
                    [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                    [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                    [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                ],
            }
        )
    )

    print(df.laktory.has_column("symbol"))
    # > False
    print(df.laktory.has_column("`stock`.`symbol`"))
    # > True
    print(df.laktory.has_column("`prices[2]`.`close`"))
    # > True
    ```
    """
    _col = re.sub(r"\[(\d+)\]", r"[*]", col)
    _col = re.sub(r"`", "", _col)
    return _col in self._df.laktory.schema_flat()



from typing import Literal

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class RecipientPropertyKvPairs(BaseModel):
    """Additional properties for recipient"""

    properties: dict[str, str] = Field(
        ...,
        description="a map of string key-value pairs with recipient's properties. Properties with name starting with databricks. are reserved.",
    )


class RecipientIpAccessList(BaseModel):
    """Allowed IP Address"""

    allowed_ip_addresses: list[str] = Field(
        None, description="Allowed IP Addresses in CIDR notation. Limit of 100."
    )


class RecipientToken(BaseModel):
    """Recipient Token"""

    activation_url: str = Field(
        None,
        description="Full activation URL to retrieve the access token. It will be empty if the token is already retrieved.",
    )
    created_at: int = Field(
        None,
        description="Time at which this recipient was created, in epoch milliseconds.",
    )
    created_by: str = Field(None, description="Username of recipient creator.")
    expiration_time: int = Field(
        None, description="Expiration timestamp of the token in epoch milliseconds."
    )
    id: str = Field(None, description="Unique ID of the recipient token.")
    updated_at: int = Field(
        None,
        description="Time at which this recipient was updated, in epoch milliseconds.",
    )
    updated_by: str = Field(None, description="Username of recipient Token updater.")


class Recipient(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Recipient for Delta Sharing

    A recipient is an entity that can receive shared data from a Databricks workspace.
    Recipients can be configured with authentication details to enable Delta Sharing.

    Examples
    --------
    ```py
    ```
    """

    name: str = Field(..., description="The name of the recipient")
    comment: str = Field(None, description="Comment about the recipient")
    owner: str = Field(None, description="The owner of the recipient")

    authentication_type: Literal["TOKEN", "DATABRICKS"] = Field(
        ..., description="The authentication type for the recipient"
    )

    data_recipient_global_metastore_id: str = Field(
        None,
        description="Global metastore id associated with the recipient. Required when `authentication_type` is DATABRICKS",
    )

    expiration_time: int = Field(
        None, description="Expiration timestamp of the token in epoch milliseconds."
    )

    ip_access_list: RecipientIpAccessList = Field(
        None, description="IP access list for the recipient"
    )

    properties_kvpairs: dict = Field(
        None, description="Additional properties as key-value pairs"
    )

    region: str = Field(
        None,
        description="Cloud region of the recipient's Unity Catalog Metastore. This field is only present when the authentication_type is `DATABRICKS`.",
    )

    sharing_code: str = Field(
        None, description="The one-time sharing code provided by the data recipient."
    )

    tokens: list[RecipientToken] = Field(
        None,
        description="List of Recipient Tokens. This field is only present when the authentication_type is TOKEN. ",
    )

    # updated_at: int = Field(None, description="Time at which this recipient was updated, in epoch milliseconds.")
    # updated_by: str = Field(None, description="Username of recipient Token updater.")

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Recipient"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_recipient"



from typing import Any
from typing import Literal

from narwhals import LazyFrame
from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.datasinks.tabledatasink import TableDataSink
from laktory.models.resources.databricks.qualitymonitor import QualityMonitor

logger = get_logger(__name__)


class UnityCatalogDataSink(TableDataSink):
    """
    Data sink writing to a Unity Catalog data table

    Examples
    ---------
    ```python tag:skip-run
    from laktory import models

    df = spark.createDataFrame([{"x": 1}, {"x": 2}, {"x": 3}])

    sink = models.UnityCatalogDataSink(
        catalog_name="dev",
        schema_name="my_schema",
        table_name="my_stable",
        mode="OVERWRITE",
    )
    sink.write(df)
    ```
    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/sourcessinks/)
    """

    type: Literal["UNITY_CATALOG"] = Field(
        "UNITY_CATALOG", frozen=True, description="Sink Type"
    )
    databricks_quality_monitor: QualityMonitor | None = Field(
        None,
        description="Databricks Quality Monitor. Requires `databricks_quality_monitor_enabled` set to `True`.",
    )

    @model_validator(mode="after")
    def validate_qm(self) -> Any:
        if self.databricks_quality_monitor is not None:
            if (
                self.parent_pipeline
                and not self.parent_pipeline.databricks_quality_monitor_enabled
            ):
                raise ValueError(
                    "`databricks_quality_monitor_enabled` must be set to `True` to use quality monitor on Unity Catalog data sinks."
                )
        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _write_polars(self, df) -> LazyFrame:
        # TODO
        # https://docs.pola.rs/api/python/stable/reference/catalog/api/polars.Catalog.scan_table.html#polars.Catalog.scan_table
        raise NotImplementedError(
            "Unity Catalog data sink is not yet implemented for Polars"
        )



from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class OboToken(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks On-Behalf-Of token
    """

    application_id: str = Field(
        None,
        description="Application ID of databricks.ServicePrincipal to create a PAT token for.",
    )
    comment: str = Field(
        None, description="Comment that describes the purpose of the token."
    )
    lifetime_seconds: str = Field(
        None,
        description="The number of seconds before the token expires. Token resource is re-created when it expires. If no lifetime is specified, the token remains valid indefinitely.",
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #
    #
    @property
    def resource_key(self) -> str:
        return self.application_id

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:OboToken"

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"value": "string_value"}

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_obo_token"

    @property
    def terraform_renames(self) -> dict[str, str]:
        return self.pulumi_renames



from typing import TYPE_CHECKING

from pydantic import Field

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.dataframe.dataframeexpr import DataFrameExpr
from laktory.models.dataframe.dataframemethod import DataFrameMethod
from laktory.models.pipelinechild import PipelineChild
from laktory.typing import AnyFrame

if TYPE_CHECKING:
    pass

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class DataFrameTransformer(BaseModel, PipelineChild):
    """
    A chain of transformations to be applied to a DataFrame. Transformations can be
    SQL- or DataFrame API-based.

    Examples
    --------
    ```py
    import polars as pl

    import laktory as lk

    df0 = pl.DataFrame(
        {
            "id": ["a", "b", "c"],
            "x1": [1, 2, 3],
        }
    )

    node0 = lk.models.DataFrameMethod(
        func_name="with_columns",
        func_kwargs={
            "y1": "x1",
        },
    )
    node1 = lk.models.DataFrameExpr(expr="select id, x1, y1 from {df}")
    transformer = lk.models.DataFrameTransformer(nodes=[node0, node1])

    df = transformer.execute(df0).collect()

    print(df)
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    | | id | x1 | y1 | |
    | |----|----|----| |
    | | a  | 1  | 1  | |
    | | b  | 2  | 2  | |
    | | c  | 3  | 3  | |
    └──────────────────┘
    '''
    ```
    References
    ----------
    * [Data Transformer](https://www.laktory.ai/concepts/transformer/)
    """

    nodes: list[DataFrameMethod | DataFrameExpr] = Field(
        ..., description="List of transformations"
    )

    @property
    def upstream_node_names(self) -> list[str]:
        """Pipeline node names required to apply transformer"""
        names = []
        for node in self.nodes:
            names += node.upstream_node_names
        return names

    @property
    def data_sources(self):
        """Get all sources feeding the Transformer"""

        sources = []
        for node in self.nodes:
            sources += node.data_sources

        return sources

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def children_names(self):
        return ["nodes"]

    # ----------------------------------------------------------------------- #
    # Execution                                                               #
    # ----------------------------------------------------------------------- #

    def execute(self, df, named_dfs=None) -> AnyFrame:
        """
        Execute transformation nodes on provided DataFrame `df`

        Parameters
        ----------
        df:
            Input dataframe
        named_dfs:
            Other DataFrame(s) to be passed to the method.

        Returns
        -------
            Output dataframe
        """
        logger.info("Executing DataFrame Transformer")

        if named_dfs is None:
            named_dfs = {}

        for inode, node in enumerate(self.nodes):
            tnode = type(node)
            logger.info(
                f"Executing DataFrame transformer node {inode} ({tnode.__name__})."
            )

            if isinstance(node, DataFrameMethod):
                df = node.execute(df)
            elif isinstance(node, DataFrameExpr):
                dfs = {}
                if df is not None:
                    dfs["df"] = df
                dfs = dfs | named_dfs
                df = node.to_df(dfs)
            else:
                raise NotImplementedError()

        return df


# BaseModel.model_rebuild()



from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class SharedObjectPartitionValue(BaseModel):
    """Shared Object Partition Value"""

    name: str = Field(..., description="The name of the partition column.")
    op: str = Field(
        ..., description="The operator to apply for the value, one of: EQUAL, LIKE"
    )
    recipient_property_key: str = Field(
        None,
        description="The key of a Delta Sharing recipient's property. For example `databricks-account-id`. When this field is set, field `value` can not be set.",
    )
    value: str = Field(
        None,
        description="The value of the partition column. When this value is not set, it means null value. When this field is set, field `recipient_property_key` can not be set.",
    )


class ShareObjectPartition(BaseModel):
    """Shared Object Partition"""

    values: list[SharedObjectPartitionValue] = Field(
        None,
        description="The value of the partition column. When this value is not set, it means null value. When this field is set, field recipient_property_key can not be set.",
    )


class ShareObject(BaseModel):
    """Object to be shared"""

    name: str = Field(..., description="Name of the object")
    data_object_type: str = Field(
        ..., description="Type of the data object (e.g., TABLE, VIEW)"
    )
    added_at: int = Field(None, description="Timestamp when object was added")
    added_by: str = Field(None, description="User who added the object")
    cdf_enabled: bool = Field(
        None,
        description="Whether to enable Change Data Feed (cdf) on the shared object. When this field is set, field `history_data_sharing_status` can not be set.",
    )
    comment: str = Field(None, description="Comment about the shared object")
    content: str = Field(None, description="")
    history_data_sharing_status: str = Field(
        None, description="History data sharing status"
    )
    partitions: list[ShareObjectPartition] = Field(
        None, description="Partition values for the shared object"
    )
    shared_as: str = Field(None, description="Name the object is shared as")
    start_version: int = Field(None, description="Start version for sharing")
    status: str = Field(None, description="Status of the object")


class ShareProviderConfig(BaseModel):
    """Provider Config"""

    workspace_id: str = Field(
        ...,
        description="Workspace ID which the resource belongs to. This workspace must be part of the account which the provider is configured with.",
    )


class Share(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Share for Delta Sharing

    A share is a container that holds the objects to be shared with recipients.
    Shares enable data sharing between Databricks workspaces.

    Examples
    --------
    ```py
    ```
    """

    name: str = Field(..., description="The name of the share")
    comment: str = Field(None, description="Comment about the share")
    owner: str = Field(None, description="The owner of the share")
    objects: list[ShareObject] = Field(
        None, description="Objects contained in the share"
    )
    provider_config: ShareProviderConfig = Field(None, description="Provider config")

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Share"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_share"



from pathlib import Path
from typing import Any
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field
from pydantic import model_validator

from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Dashboard(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Lakeview Dashboard

    Examples
    --------
    ```py
    import io

    from laktory import models

    # Define job
    job_yaml = '''
    display_name: databricks-costs
    file_path: ./dashboards/databricks_costs.json
    parent_path: /.laktory/dashboards
    warehouse_id: a7d9f2kl8mp3q6rt
    access_controls:
        - group_name: account users
          permission_level: CAN_READ
        - group_name: account users
          permission_level: CAN_RUN
    '''
    dashboard = models.resources.databricks.Dashboard.model_validate_yaml(
        io.StringIO(job_yaml)
    )
    ```
    """

    access_controls: list[AccessControl] = Field([], description="Access controls list")
    display_name: str = Field(..., description="The display name of the dashboard.")
    embed_credentials: bool = Field(
        None,
        description="Whether to embed credentials in the dashboard. Default is true.",
    )
    file_path: str = Field(
        None,
        description="The path to the dashboard JSON file. Conflicts with serialized_dashboard.",
    )
    name_prefix: str = Field(
        None, description="Prefix added to the dashboard display name"
    )
    name_suffix: str = Field(
        None, description="Suffix added to the dashboard display name"
    )
    parent_path_: str | None = Field(
        None,
        description="""
        The path to a workspace folder (inside laktory root) containing the dashboard. If changed, the query will be
        recreated.
        """,
        validation_alias=AliasChoices("parent_path", "dirpath", "parent_path_"),
        exclude=True,
    )
    serialized_dashboard: str = Field(
        None,
        description="The contents of the dashboard in serialized string form. Conflicts with file_path.",
    )
    warehouse_id: str = Field(
        ..., description="The warehouse ID used to run the dashboard."
    )

    @computed_field(description="parent_path")
    @property
    def parent_path(self) -> str:
        if self.parent_path_ is None:
            self.parent_path_ = ""
        if self.parent_path_.startswith("/"):
            self.parent_path_ = self.parent_path_[1:]

        parent_path = Path(settings.workspace_laktory_root) / self.parent_path_
        return parent_path.as_posix()

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.display_name = self.name_prefix + self.display_name
                self.name_prefix = ""
            if self.name_suffix:
                self.display_name = self.display_name + self.name_suffix
                self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.display_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    dashboard_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Dashboard"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "access_controls",
            "name_prefix",
            "name_suffix",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_dashboard"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import laktory as lk
from laktory import models

# --------------------------------------------------------------------------- #
# Code Pipeline                                                               #
# --------------------------------------------------------------------------- #

# While one of the main benefit of using Laktory is the ability to declare
# pipeline from a serialized format, there is nothing prevent from creating
# them directly in the code. Here is an example.

# --------------------------------------------------------------------------- #
# Bronze Node                                                                 #
# --------------------------------------------------------------------------- #

node_brz = models.PipelineNode(
    name="brz_stock_prices",
    source=models.FileDataSource(
        path="./data/stock_prices.json",
        format="JSONL",
    ),
    sinks=[
        models.FileDataSink(
            path="./data/brz_stock_prices.parquet",
            format="PARQUET",
        )
    ],
)


# --------------------------------------------------------------------------- #
# Silver Node                                                                 #
# --------------------------------------------------------------------------- #

node_slv = models.PipelineNode(
    name="slv_stock_prices",
    source=models.PipelineNodeDataSource(
        node_name="brz_stock_prices",
    ),
    sinks=[
        models.FileDataSink(
            path="./data/slv_stock_prices.parquet",
            format="PARQUET",
        )
    ],
    transformer=models.DataFrameTransformer(
        nodes=[
            models.DataFrameExpr(
                expr="""
                    SELECT
                      CAST(data.created_at AS TIMESTAMP) AS created_at,
                      data.symbol AS name,
                      data.symbol AS symbol,
                      data.open AS open,
                      data.close AS close,
                      data.high AS high,
                      data.low AS low,
                      data.volume AS volume
                    FROM
                      {df}                
                """
            ),
            models.DataFrameMethod(
                func_name="unique",
                func_kwargs={
                    "subset": ["symbol", "created_at"],
                    "keep": "any",
                },
            ),
        ]
    ),
)


# --------------------------------------------------------------------------- #
# Gold Node                                                                   #
# --------------------------------------------------------------------------- #


@lk.api.register_anyframe_namespace("custom")
class CustomNamespace:
    def __init__(self, _df):
        self._df = _df

    def process_stocks(self):
        import polars as pl

        df = self._df.to_native()  # convert from Narwhals to Polars

        df = df.with_columns(
            open_rounder=pl.col("open").round(2),
            close_rounded=pl.col("close").round(2),
        )

        return df


node_gld = models.PipelineNode(
    name="gld_stock_prices",
    source=models.PipelineNodeDataSource(
        node_name="slv_stock_prices",
    ),
    transformer=models.DataFrameTransformer(
        nodes=[
            models.DataFrameMethod(
                func_name="custom.process_stocks",
            ),
        ]
    ),
)


# --------------------------------------------------------------------------- #
# Execute Pipeline                                                            #
# --------------------------------------------------------------------------- #

pipeline = models.Pipeline(
    name="pl-stock-prices",
    dataframe_backend="POLARS",
    nodes=[
        node_brz,
        node_slv,
        node_gld,
    ],
)

# Run
pipeline.execute()



import os
from pathlib import Path

import narwhals as nw
import pandas as pd
import polars as pl

from laktory import get_spark_session
from laktory.enums import DataFrameBackends


def get_backend(v):
    if isinstance(v, str):
        return DataFrameBackends(v)
    return DataFrameBackends.from_nw_implementation(nw.from_native(v).implementation)


def to_backend(df, backend, lazy=False):
    backend = get_backend(backend)

    if backend == DataFrameBackends.POLARS:
        df = pl.from_pandas(df)
        if lazy:
            df = df.lazy()
    elif backend == DataFrameBackends.PYSPARK:
        spark = get_spark_session()
        df = spark.createDataFrame(df)
    return nw.from_native(df)


def get_df0(backend, lazy=False):
    df = pd.DataFrame(
        {
            "_idx": [0, 1, 2],
            "id": ["a", "b", "c"],
            "x1": [1, 2, 3],
        }
    )

    return to_backend(df, backend, lazy=lazy)


def get_df1(backend, lazy=False):
    df = pd.DataFrame(
        {
            "_idx": [0, 1, 2],
            "id": ["b", "c", "d"],
            "x2": [4, 9, 16],
        }
    )

    return to_backend(df, backend, lazy=lazy)


class StreamingSource:
    def __init__(
        self,
        backend="POLARS",
    ):
        self.backend = backend
        self.ibatch = 0
        self.irow = -1

    def get_dfs(self, nbatch=1):
        dfs = []
        for _ibatch in range(nbatch):
            df = get_df0(self.backend, lazy=True)
            df = df.with_columns(_batch_id=nw.lit(self.ibatch))
            df = df.with_columns(_idx=nw.col("_idx") + self.ibatch * 3)
            dfs += [df]
            self.ibatch += 1
        return dfs

    def write_to_delta(self, filepath, nbatch=1):
        is_init = self.ibatch == 0

        for df in self.get_dfs(nbatch):
            if self.backend == "PYSPARK":
                if is_init:
                    df.to_native().write.format("DELTA").mode("overwrite").save(
                        filepath
                    )
                else:
                    df.to_native().write.format("DELTA").mode("append").save(filepath)
            else:
                raise NotImplementedError()

    def write_to_json(self, dirpath, nbatch=1):
        dirpath = Path(dirpath)

        for df in self.get_dfs(nbatch):
            _df = df.collect().to_pandas()

            ibatch = _df["_batch_id"].max()

            if not dirpath.is_dir():
                os.mkdir(dirpath)

            filepath = Path(dirpath) / f"{ibatch:03d}.json"
            filepath = str(filepath.absolute())

            _df.to_json(filepath, orient="records")

        return df



from typing import Literal
from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.grants.storagecredentialgrant import StorageCredentialGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class AwsIamRole(BaseModel):
    role_arn: str = Field(None, description="")
    external_id: str = Field(None, description="")
    unity_catalog_iam_arn: str = Field(None, description="")


class AzureManagedIdentity(BaseModel):
    access_connector_id: str = Field(None, description="")
    credential_id: str = Field(None, description="")
    managed_identity_id: str = Field(None, description="")


class AzureServicePrincipal(BaseModel):
    application_id: str = Field(None, description="")
    client_secret: str = Field(None, description="")
    directory_id: str = Field(None, description="")


class CloudflareApiToken(BaseModel):
    account_id: str = Field(None, description="R2 account ID")
    access_key_id: str = Field(None, description="R2 API token access key ID")
    secret_access_key: str = Field(None, description="R2 API token secret access key")


class DatabricksGcpServiceAccount(BaseModel):
    credential_id: str = Field(None, description="")
    email: str = Field(None, description="")


class GcpServiceAccountKey(BaseModel):
    email: str = Field(None, description="")
    private_key: str = Field(None, description="")
    private_key_id: str = Field(None, description="")


class StorageCredential(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Storage Credential

    Examples
    --------
    ```py
    ```
    """

    aws_iam_role: AwsIamRole = Field(None, description="AWS IAM role specifications")
    azure_managed_identity: AzureManagedIdentity = Field(
        None, description="Azure Managed Identity specifications"
    )
    azure_service_principal: AzureServicePrincipal = Field(
        None, description="Azure Service Principal specifications"
    )
    cloudflare_api_token: CloudflareApiToken = Field(
        None,
        description="""
        Optional configuration block for using a Cloudflare API Token as credential details. 
        This requires account admin access.
        """,
    )
    comment: str = Field(None, description="Comment")
    databricks_gcp_service_account: DatabricksGcpServiceAccount = Field(
        None, description="Databricks GCP service account specifications"
    )
    force_destroy: bool = Field(
        None, description="Force resource deletion even if not empty"
    )
    force_update: bool = Field(
        None, description="Force resource update even if not empty"
    )
    gcp_service_account_key: GcpServiceAccountKey = Field(None, description="")
    grant: Union[StorageCredentialGrant, list[StorageCredentialGrant]] = Field(
        None,
        description="""
    Grant(s) operating on the Storage Credential and authoritative for a specific principal. Other principals within 
    the grants are preserved. Mutually exclusive with `grants`.
    """,
    )
    grants: list[StorageCredentialGrant] = Field(
        None,
        description="""
    Grants operating on the Storage Credential and authoritative for all principals. Replaces any existing grants 
    defined inside or outside of Laktory. Mutually exclusive with `grant`.
    """,
    )
    isolation_mode: Literal["ISOLATION_MODE_ISOLATED", "ISOLATION_MODE_OPEN"] = Field(
        None, description="GCP service account key specifications"
    )
    metastore_id: str = Field(None, description="Metastore id")
    name: str = Field(None, description="Data Access name")
    owner: str = Field(None, description="Owner")
    read_only: bool = Field(None, description="Read only")
    skip_validation: bool = Field(None, description="Skip Validation")

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - storage credential grants
        """
        resources = []

        # Storage Credential grants
        resources += self.get_grants_additional_resources(
            object={"storage_credential": f"${{resources.{self.resource_name}.id}}"}
        )
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:StorageCredential"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["grant", "grants"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_storage_credential"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from pathlib import Path
from typing import Any
from typing import Literal
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field
from pydantic import model_validator

from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.alert import Alert
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class QueryParameterTextValue(BaseModel):
    value: str = Field(..., description="Actual text value.")


class QueryParameterQueryBackedValueMultiValuesOptions(BaseModel):
    prefix: str = Field(
        ..., description="Character that prefixes each selected parameter value."
    )
    separator: str = Field(
        ...,
        description="Character that separates each selected parameter value. Defaults to a comma.",
    )
    suffix: str = Field(
        ..., description="Character that suffixes each selected parameter value."
    )


class QueryParameterQueryBackedValue(BaseModel):
    query_id: str = Field(
        None, description="ID of the query that provides the parameter values."
    )
    multi_values_options: QueryParameterQueryBackedValueMultiValuesOptions = Field(
        None,
        description="If specified, allows multiple values to be selected for this parameter.",
    )
    values: list[str] = Field(
        None, description="List of selected query parameter values."
    )


class QueryParameterNumericValue(BaseModel):
    value: float = Field(None, description="Actual numeric value")


class QueryParameterEnumValueMultiValuesOptions(BaseModel):
    prefix: str = Field(
        None, description="Character that prefixes each selected parameter value."
    )
    separator: str = Field(
        None,
        description="Character that separates each selected parameter value. Defaults to a comma.",
    )
    suffix: str = Field(
        None, description="Character that suffixes each selected parameter value."
    )


class QueryParameterEnumValue(BaseModel):
    enum_options: str = Field(
        None, description="List of valid query parameter values, newline delimited."
    )
    multi_values_options: QueryParameterEnumValueMultiValuesOptions = Field(
        None,
        description="If specified, allows multiple values to be selected for this parameter.",
    )
    values: list[str] = Field(
        None, description="List of selected query parameter values."
    )


class QueryParameterDateValue(BaseModel):
    date_value: str = Field(None, description="Manually specified date-time value")
    dynamic_date_value: Literal["NOW", "YESTERDAY"] = Field(
        None, description="Dynamic date-time value based on current date-time"
    )
    precision: Literal["DAY_PRECISION", "MINUTE_PRECISION", "SECOND_PRECISION"] = Field(
        None,
        description="Date-time precision to format the value into when the query is run.",
    )


class QueryParameterDateRangeValueDateRangeValue(BaseModel):
    end: str = Field(..., description="end of the date range")
    start: str = Field(..., description="start of the date range")


class QueryParameterDateRangeValue(BaseModel):
    date_range_value: QueryParameterDateRangeValueDateRangeValue = Field(
        None, description="Manually specified date-time range value"
    )
    dynamic_date_range_value: Literal[
        "TODAY",
        "YESTERDAY",
        "THIS_WEEK",
        "THIS_MONTH",
        "THIS_YEAR",
        "LAST_WEEK",
        "LAST_MONTH",
        "LAST_YEAR",
        "LAST_HOUR",
        "LAST_8_HOURS",
        "LAST_24_HOURS",
        "LAST_7_DAYS",
        "LAST_14_DAYS",
        "LAST_30_DAYS",
        "LAST_60_DAYS",
        "LAST_90_DAYS",
        "LAST_12_MONTHS",
    ] = Field(
        None, description="Dynamic date-time range value based on current date-time."
    )
    precision: str = Field(None, description="")
    start_day_of_week: str = Field(None, description="")


class QueryParameter(BaseModel):
    name: str = Field(
        ...,
        description="Literal parameter marker that appears between double curly braces in the query text.",
    )
    date_range_value: QueryParameterDateRangeValue = Field(
        None,
        description="""
    Date-range query parameter value. Consists of following attributes (Can only specify one of 
    `dynamic_date_range_value` or`date_range_value`)
    """,
    )
    date_value: QueryParameterDateValue = Field(
        None,
        description="""
        Date query parameter value. Consists of following attributes (Can only specify one of 
        `dynamic_date_value` or `date_value`)
        """,
    )
    enum_value: QueryParameterEnumValue = Field(
        None, description="Dropdown parameter value"
    )
    numeric_value: QueryParameterNumericValue = Field(
        None, description="Numeric parameter value"
    )
    query_backed_value: QueryParameterQueryBackedValue = Field(
        None, description="Query-based dropdown parameter value"
    )
    text_value: QueryParameterTextValue = Field(
        None, description="Text parameter value"
    )
    title: str = Field(
        None,
        description="Text displayed in the user-facing parameter widget in the UI.",
    )


class Query(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Query

    Examples
    --------
    ```py
    from laktory import models

    query = models.resources.databricks.Query(
        display_name="google-prices",
        parent_path="/queries",
        query_text="SELECT * FROM dev.finance.slv_stock_prices",
        warehouse_id="12345",
    )
    ```
    """

    access_controls: list[AccessControl] = Field([], description="Access controls list")
    alert: Alert = Field(None, description="")
    apply_auto_limit: bool = Field(
        None, description="Whether to apply a 1000 row limit to the query result."
    )
    catalog: str = Field(
        None, description="Name of the catalog where this query will be executed."
    )
    description: str = Field(
        None,
        description="General description that conveys additional information about this query such as usage notes.",
    )
    display_name: str = Field(..., description="Name of the query.")
    name_prefix: str = Field(None, description="")
    name_suffix: str = Field(None, description="")
    owner_user_name: str = Field(None, description="Query owner's username.")
    parameters: list[QueryParameter] = Field(
        None,
        description="Query parameter definition. Consists of following attributes (one of `*_value` is required):",
    )
    parent_path_: str | None = Field(
        None,
        description="""
        The path to a workspace folder (inside laktory root) containing the query. If changed, the query will be
        recreated.
        """,
        validation_alias=AliasChoices("parent_path", "dirpath", "parent_path_"),
        exclude=True,
    )
    query_text: str = Field(..., description="Text of SQL query.")
    run_as_mode: str = Field(None, description="Sets the 'Run as' role for the object.")
    schema_: str = Field(
        None,
        validation_alias=AliasChoices("schema", "schema_"),
        description="Name of the schema where this query will be executed.",
    )  # required not to overwrite BaseModel attribute
    tags: list[str] = Field(None, description="Tags that will be added to the query.")
    warehouse_id: str = Field(
        ...,
        description="ID of a SQL warehouse which will be used to execute this query.",
    )

    @computed_field(description="parent_path")
    @property
    def parent_path(self) -> str:
        if self.parent_path_ is None:
            self.parent_path_ = ""
        if self.parent_path_.startswith("/"):
            self.parent_path_ = self.parent_path_[1:]

        parent_path = Path(settings.workspace_laktory_root) / self.parent_path_
        return parent_path.as_posix()

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.display_name = self.name_prefix + self.display_name
                self.name_prefix = ""
            if self.name_suffix:
                self.display_name = self.display_name + self.name_suffix
                self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.display_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        - alert
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    sql_query_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        if self.alert:
            self.alert.query_id = f"${{resources.{self.resource_name}.id}}"
            resources += [self.alert]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"schema_": "schema"}

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Query"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "access_controls",
            "alert",
            "name_prefix",
            "name_suffix",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_query"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_renames(self) -> dict[str, str]:
        return self.pulumi_renames



import json
from typing import Literal

from pydantic import Field

from laktory.models.datasinks.tabledatasink import TableDataSink
from laktory.models.pipeline.orchestrators.pipelineconfigworkspacefile import (
    PipelineConfigWorkspaceFile,
)
from laktory.models.pipelinechild import PipelineChild
from laktory.models.resources.databricks.pipeline import Pipeline
from laktory.models.resources.pulumiresource import PulumiResource


class DatabricksPipelineOrchestrator(Pipeline, PipelineChild):
    """
    Databricks Pipeline used as an orchestrator to execute a Laktory pipeline.

    DLT orchestrator does not support pipeline nodes with views (as opposed to
    materialized tables). Also, it does not support writing to multiple
    schemas within the same pipeline.

    Selecting this orchestrator requires to add the supporting
    [notebook](https://github.com/okube-ai/laktory/blob/main/laktory/resources/quickstart-stacks/workflows/notebooks/dlt/dlt_laktory_pl.py)
    to the stack.


    References
    ----------
    * [Data Pipeline](https://www.laktory.ai/concepts/pipeline/)
    * [Databricks DLT](https://www.databricks.com/product/delta-live-tables)
    """

    type: Literal["DATABRICKS_PIPELINE"] = Field(
        "DATABRICKS_PIPELINE", description="Type of orchestrator"
    )
    config_file: PipelineConfigWorkspaceFile = Field(
        PipelineConfigWorkspaceFile(),
        description="Pipeline configuration (json) file deployed to the workspace and used by the job to read and execute the pipeline.",
    )

    # ----------------------------------------------------------------------- #
    # Update DLT                                                              #
    # ----------------------------------------------------------------------- #

    def update_from_parent(self):
        pl = self.parent_pipeline

        for node in pl.nodes:
            if node.is_view:
                raise ValueError(
                    f"Node '{node.name}' of pipeline '{pl.name}' is a view which is not supported with DLT orchestrator."
                )

        for n in pl.nodes:
            for s in n.all_sinks:
                if isinstance(s, TableDataSink):
                    s.catalog_name = s.catalog_name or self.catalog
                    s.schema_name = s.schema_name or self.schema_ or self.target

        # Update pipeline config
        _requirements = self.inject_vars_into_dump({"deps": pl._dependencies})["deps"]
        _path = (
            "/Workspace"
            + self.inject_vars_into_dump({"path": self.config_file.path})["path"]
        )
        self.configuration["pipeline_name"] = pl.name  # only for reference
        self.configuration["requirements"] = json.dumps(_requirements)
        self.configuration["config_filepath"] = _path
        # This is to ensure configuration is flagged as set and part of
        # model_fields_set when injecting variables.
        self.configuration = self.configuration

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def children_names(self):
        return ["config_file", "type"]

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        return "dlt-pipeline"

    @property
    def pulumi_excludes(self) -> list[str] | dict[str, bool]:
        excludes = super().pulumi_excludes
        excludes["config_file"] = True
        excludes["type"] = True
        excludes["dataframe_backend"] = True
        excludes["dataframe_api"] = True

        return excludes

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - configuration workspace file
        - configuration workspace file permissions
        """

        resources = super().additional_core_resources
        resources += [self.config_file]

        return resources
